part of pdf;

/// PDFCrossTable is responsible for intermediate level parsing and savingof a PDF document.
class _PdfCrossTable {
  //Constructor
  _PdfCrossTable();

  //Fields
  PdfDocument _pdfDocument;
  int _count = 0;
  _PdfMainObjectCollection _items;
  _PdfDictionary _trailer;
  Map<int, _RegisteredObject> _objects = <int, _RegisteredObject>{};

  //Properties
  int get nextObjectNumber {
    _count = _count + 1;
    return _count;
  }

  _PdfDictionary get trailer {
    _trailer ??= _PdfStream();
    return _trailer;
  }

  _PdfMainObjectCollection get _objectCollection => _pdfDocument._objects;
  PdfDocument get _document => _pdfDocument;
  set _document(PdfDocument document) {
    if (document == null) {
      throw ArgumentError('Document');
    }
    _pdfDocument = document;
    _items = _pdfDocument._objects;
  }

  //Implementation
  void _save(_PdfWriter writer) {
    ArgumentError.checkNotNull(writer, 'writer');
    _saveHead(writer);
    _saveObjects(writer);
    final int saveCount = _count;
    final int referencePosition = writer._position;
    _registerObject(0, _PdfReference(0, -1), true);
    writer._write(_Operators.crossReference);
    writer._write(_Operators.newLine);
    _saveSections(writer);
    _saveTrailer(writer, _count);
    _saveEnd(writer, referencePosition);
    _count = saveCount;
    for (int i = 0; i < _objectCollection._count; i++) {
      final _ObjectInfo objectInfo = _objectCollection[i];
      objectInfo._object.isSaving = false;
    }
  }

  void _saveHead(_PdfWriter writer) {
    writer._write('%PDF-');
    final String version = _generateFileVersion(writer._document);
    writer._write(version);
    writer._write(_Operators.newLine);
    writer._write(<int>[0x25, 0x83, 0x92, 0xfa, 0xfe]);
    writer._write(_Operators.newLine);
  }

  String _generateFileVersion(PdfDocument document) {
    if (document.fileStructure.version == PdfVersion.version2_0) {
      const String version = '2.0';
      return version;
    } else {
      return '1.' + document.fileStructure.version.index.toString();
    }
  }

  void _saveObjects(_PdfWriter writer) {
    final _PdfMainObjectCollection objectCollection = _objectCollection;
    for (int i = 0; i < objectCollection._count; i++) {
      final _ObjectInfo objInfo = objectCollection[i];
      final _IPdfPrimitive obj = objInfo._object;
      obj.isSaving = true;
      _saveIndirectObject(obj, writer);
    }
  }

  void _saveIndirectObject(_IPdfPrimitive object, _PdfWriter writer) {
    ArgumentError.checkNotNull(object, 'object');
    ArgumentError.checkNotNull(writer, 'writer');
    final _PdfReference reference = _getReference(object);
    if (object is _PdfCatalog) {
      trailer[_DictionaryProperties.root] = reference;
    }
    _registerObject(writer._position, reference);
    _doSaveObject(object, reference, writer);
  }

  _PdfReference _getReference(_IPdfPrimitive object) {
    bool wasNew = false;
    if (object is _PdfReferenceHolder) {
      object = (object as _PdfReferenceHolder).object;
      object.isSaving = true;
    }
    if (object is _IPdfWrapper) {
      object = (object as _IPdfWrapper)._element;
    }
    dynamic reference;
    if (_items._count > 0 &&
        object.objectCollectionIndex > 0 &&
        _items._count > object.objectCollectionIndex - 1) {
      reference = _document._objects._getReference(object);
    }
    if (reference == null) {
      if (object.status == _ObjectStatus.registered) {
        wasNew = false;
      } else {
        wasNew = true;
      }
    } else {
      wasNew = false;
    }
    if (reference == null) {
      final int objectNumber = nextObjectNumber;
      reference = _PdfReference(objectNumber, 0);
      if (wasNew) {
        _document._objects._add(object);
        _document._objects._trySetReference(object, reference);
        final int tempIndex = _document._objects._count - 1;
        final int tempkey =
            _document._objects._objectCollection[tempIndex]._reference._objNum;
        final _ObjectInfo tempvalue =
            _document._objects._objectCollection[_document._objects._count - 1];
        _document._objects._mainObjectCollection[tempkey] = tempvalue;
        object.position = -1;
      } else {
        _document._objects._trySetReference(object, reference);
      }
      object.objectCollectionIndex = reference._objNum;
      object.status = _ObjectStatus.none;
    }
    return reference;
  }

  void _registerObject(int position, _PdfReference reference,
      [bool isFreeType]) {
    ArgumentError.checkNotNull(reference, 'reference');
    if (isFreeType == null) {
      _objects[reference._objNum] = _RegisteredObject(position, reference);
    } else {
      _objects[reference._objNum] =
          _RegisteredObject(position, reference, isFreeType);
    }
  }

  void _doSaveObject(
      _IPdfPrimitive object, _PdfReference reference, _PdfWriter writer) {
    writer._write(reference._objNum);
    writer._write(_Operators.whiteSpace);
    writer._write(reference._genNum);
    writer._write(_Operators.whiteSpace);
    writer._write(_Operators.obj);
    writer._write(_Operators.newLine);
    object.save(writer);
    if (object is _PdfName || object is _PdfNumber || object is _PdfNull) {
      writer._write(_Operators.newLine);
    }
    writer._write(_Operators.endobj);
    writer._write(_Operators.newLine);
  }

  void _saveSections(_IPdfWriter writer) {
    int objectNumber = 0;
    int count = 0;
    do {
      count = _prepareSubsection(objectNumber);
      _saveSubsection(writer, objectNumber, count);
      objectNumber += count;
    } while (count != 0);
  }

  int _prepareSubsection(int objectNumber) {
    int count = 0;
    int i;
    final int total = _count;
    if (objectNumber >= total) {
      return count;
    }
    for (i = objectNumber; i < total; ++i) {
      break;
    }
    objectNumber = i;
    for (; i < total; ++i) {
      ++count;
    }
    return count;
  }

  void _saveSubsection(_IPdfWriter writer, int objectNumber, int count) {
    if (count <= 0 || objectNumber >= _count) {
      return;
    }
    writer._write(objectNumber.toString() +
        ' ' +
        (count + 1).toString() +
        _Operators.newLine);
    for (int i = objectNumber; i <= objectNumber + count; ++i) {
      final _RegisteredObject obj = _objects[i];
      String value = '';
      if (obj._type == _ObjectType.free) {
        value = _getItem(obj._offset, 65535, true);
      } else {
        value = _getItem(obj._offset, obj._generationNumber, false);
      }
      writer._write(value);
    }
  }

  String _getItem(int offset, int generationNumber, bool isFreeType) {
    String result = '';
    final int offsetLength = 10 - offset.toString().length;
    if (generationNumber <= 0) {
      generationNumber = 0;
    }
    final int generationNumberLength =
        (5 - generationNumber.toString().length) <= 0
            ? 0
            : (5 - generationNumber.toString().length);
    for (int index = 0; index < offsetLength; index++) {
      result = result + '0';
    }
    result = result + offset.toString() + ' ';
    for (int index = 0; index < generationNumberLength; index++) {
      result = result + '0';
    }
    result = result + generationNumber.toString() + ' ';
    result = result +
        (isFreeType ? _Operators.f : _Operators.n) +
        _Operators.newLine;
    return result;
  }

  void _saveTrailer(_IPdfWriter writer, int count) {
    ArgumentError.checkNotNull(writer, 'writer');
    writer._write(_Operators.trailer);
    writer._write(_Operators.newLine);
    _PdfDictionary trailerDictionary = trailer;
    trailerDictionary[_DictionaryProperties.size] = _PdfNumber(count + 1);
    trailerDictionary = _PdfDictionary(trailerDictionary);
    trailerDictionary.save(writer);
  }

  void _saveEnd(_IPdfWriter writer, int position) {
    writer._write(_Operators.newLine +
        _Operators.startCrossReference +
        _Operators.newLine);
    writer._write(position.toString() + _Operators.newLine);
    writer._write(_Operators.endOfFileMarker);
  }

  static _IPdfPrimitive _dereference(_IPdfPrimitive element) {
    if (element != null && element is _PdfReferenceHolder) {
      final _PdfReferenceHolder holder = element;
      if (holder != null) {
        return holder.object;
      }
    }
    return element;
  }

  void _dispose() {
    if (_items != null) {
      _items._dispose();
      _items = null;
    }
    if (_objects != null && _objects.isNotEmpty) {
      _objects.clear();
      _objects = null;
    }
  }
}

/// Represents a registered object.
class _RegisteredObject {
  _RegisteredObject(int offset, _PdfReference reference, [bool isFreeType]) {
    if (reference == null) {
      throw ArgumentError.notNull('reference');
    } else {
      _offset = offset;
      _generationNumber = reference._genNum;
      if (isFreeType != null) {
        _type = isFreeType ? _ObjectType.free : _ObjectType.normal;
      } else {
        _type = _ObjectType.normal;
      }
    }
  }
  int _offset;
  int _generationNumber;
  _ObjectType _type;
}
